/*=====================================================================*/
/*    serrano/prgm/utils/icccmpanel/cron.c                             */
/*    -------------------------------------------------------------    */
/*    Author      :  Manuel Serrano                                    */
/*    Creation    :  Fri Jul 23 22:15:38 2004                          */
/*    Last change :  Sat Jul 31 06:38:11 2004 (serrano)                */
/*    Copyright   :  2004 Manuel Serrano                               */
/*    -------------------------------------------------------------    */
/*    The cron                                                         */
/*=====================================================================*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <values.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#ifdef HAVE_XPM
#  include <X11/xpm.h>
#endif

#include "icccmpanel.h"
#include "cron.h"

/*---------------------------------------------------------------------*/
/*    padding                                                          */
/*---------------------------------------------------------------------*/
#define CRON_IMAGE_PADDING 2

/*---------------------------------------------------------------------*/
/*    cron_t                                                           */
/*---------------------------------------------------------------------*/
typedef struct ipcron {
   area_t area;
   int sec;
   int oldtime;
   char *cmd;
   char mode;
   char *text;
   Pixmap icon, mask;
} ipcron_t;

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    refresh_text_cron ...                                            */
/*---------------------------------------------------------------------*/
static void
refresh_text_cron( area_t *ar ) {
   ipcron_t *ip = (ipcron_t *)ar;
   taskbar_t *tbar = ar->taskbar;
   Xinfo_t *xinfo = tbar->xinfo;
   char *str = ip->text;
   int len = strlen( str );
   int text_w = ar->width;
   int text_x, text_y;

   while( (XTextWidth( xinfo->xfsb, str, len ) >= (text_w - 2)) &&
	  (len > 0) )
      len--;

   text_x = (text_w - XTextWidth( xinfo->xfsb, str, len )) / 2;
   text_y = xinfo->xfsb->ascent + ((ar->height - xinfo->xfsb->ascent) / 3);

   draw_gradient( tbar->xinfo, ar->win,
		  0, 0,
		  ar->width,
		  ar->height - 2 - (2 * tbar->aborder),
		  0, GREY12, 0, 1 );
   
   draw_text( tbar->xinfo, ar->win, text_x, text_y, str, len, 0, BLACK );
   
   draw_relief( tbar->xinfo, ar->win, 
		0, 0,
		ar->width - 1, ar->height - 1,
		0, WHITE, GREY9, tbar->aborder );
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    refresh_image_cron ...                                           */
/*---------------------------------------------------------------------*/
static void
refresh_image_cron( area_t *ar ) {
   ipcron_t *ip = (ipcron_t *)ar;
   taskbar_t *tbar = ar->taskbar;
   Xinfo_t *xinfo = tbar->xinfo;
   int icon_x = tbar->border + CRON_IMAGE_PADDING;
   int icon_y = (tbar->height - ar->height) / 2;

   draw_gradient( tbar->xinfo, ar->win,
		  0, 0,
		  ar->width,
		  ar->height - 2 - (2 * tbar->aborder),
		  0, GREY12, 0, 1 );

   draw_pixmap( xinfo, ar->win, ip->icon, ip->mask,
		icon_x, icon_y, ar->width, ar->height );
   
   draw_relief( tbar->xinfo, ar->win, 
		0, 0,
		ar->width - 1, ar->height - 1,
		0, WHITE, GREY9, tbar->aborder );
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    button_press_cron ...                                            */
/*---------------------------------------------------------------------*/
static void
button_press_cron( XEvent *ev, area_t *ar ) {
   taskbar_t *tbar = ar->taskbar;
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    enter_notify_cron ...                                            */
/*---------------------------------------------------------------------*/
static void
enter_notify_cron( XEvent *ev, area_t *ar ) {
   taskbar_t *tbar = ar->taskbar;
   tooltips_setup( ((ipcron_t *)ar)->cmd,
		   ar->x,
		   tbar->top ? tbar->y + tbar->height : tbar->y - tbar->height,
		   TOOLTIPS );
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    leave_notify_cron ...                                            */
/*---------------------------------------------------------------------*/
static void
leave_notify_cron( XEvent *ev, area_t *ar ) {
   tooltips_hide();
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    set_cron ...                                                     */
/*---------------------------------------------------------------------*/
static void
set_cron( area_t *ar ) {
   time_t now = time( 0 );

   if( (now - ((ipcron_t *)ar)->oldtime) >= 60 ) {
      ((ipcron_t *)ar)->oldtime = now;
   }
}
   
/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    timeout_text_cron ...                                            */
/*---------------------------------------------------------------------*/
static int
timeout_text_cron( area_t *ar ) {
   set_cron( ar );
   refresh_text_cron( ar );

   return 1;
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    timeout_image_cron ...                                           */
/*---------------------------------------------------------------------*/
static int
timeout_image_cron( area_t *ar ) {
   set_cron( ar );
   refresh_text_cron( ar );

   return 1;
}

/*---------------------------------------------------------------------*/
/*    area_t *                                                         */
/*    make_cron ...                                                    */
/*---------------------------------------------------------------------*/
area_t *
make_cron( taskbar_t *tbar, int width, int height, char *cmd, int sec, char mode ) {
   ipcron_t *cl = calloc( 1, sizeof( ipcron_t ) );
   area_t *ar = &(cl->area);

   if( !ar ) exit( 10 );

   /* initialize the cron */
   ar->win = make_area_window( tbar );
   ar->name = "cron";
   
   ar->uwidth = width;
   ar->uheight = height;

   cl->sec = -1;
   
   /* register the timeout */
   set_cron( ar );
   evloop_timeout( ar );
   
   /* bind the area in the taskbar */
   ar->taskbar = tbar;
   tbar->areas = cons( ar, tbar->areas );

   switch( mode ) {
      case CRON_TEXT:
	 ar->refresh = &refresh_text_cron;
	 ar->desktop_notify = &refresh_text_cron;
	 ar->timeout = &timeout_text_cron;
	 break;
      case CRON_IMAGE:
	 ar->refresh = &refresh_image_cron;
	 ar->desktop_notify = &refresh_image_cron;
	 ar->timeout = &timeout_image_cron;
	 break;
      default:
	 fprintf( stderr, "Illegal cron mode: %d\n", mode );
	 ar->refresh = &refresh_text_cron;
	 ar->desktop_notify = &refresh_text_cron;
   }
   
   ar->button_press = &button_press_cron;
   ar->enter_notify = &enter_notify_cron;
   ar->leave_notify = &leave_notify_cron;

   return ar;
}
