/*=====================================================================*/
/*    serrano/prgm/utils/icccmpanel/evloop.c                           */
/*    -------------------------------------------------------------    */
/*    Author      :  Manuel Serrano                                    */
/*    Creation    :  Fri Jul 23 05:59:11 2004                          */
/*    Last change :  Fri Jul 30 15:48:11 2004 (serrano)                */
/*    Copyright   :  2004 Manuel Serrano                               */
/*    -------------------------------------------------------------    */
/*    Event loop                                                       */
/*=====================================================================*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <values.h>

#ifdef HAVE_XPM
#  include <X11/xpm.h>
#endif

#include "icccmpanel.h"

/*---------------------------------------------------------------------*/
/*    static pair_t *                                                  */
/*    timeout_areas ...                                                */
/*---------------------------------------------------------------------*/
static pair_t *timeout_areas = NIL;

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    evloop_timeout ...                                               */
/*---------------------------------------------------------------------*/
void
evloop_timeout( area_t *ar ) {
   timeout_areas = cons( ar, timeout_areas );
}

/*---------------------------------------------------------------------*/
/*    static area_t *                                                  */
/*    enter_area ...                                                   */
/*---------------------------------------------------------------------*/
static area_t *
enter_area( XEvent *ev, area_t *newa, area_t *olda ) {
   if( newa != olda ) {
      if( olda && olda->leave_notify ) olda->leave_notify( ev, olda );
      if( newa && newa->enter_notify ) newa->enter_notify( ev, newa );
   }
   return newa;
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    evloop ...                                                       */
/*---------------------------------------------------------------------*/
void
evloop( taskbar_t *tbar ) {
   area_t *iar = 0;
   XEvent ev;
   int evt;
   XEvent elev;
   fd_set fd;
   int xfd;
   
   xfd = ConnectionNumber( tbar->xinfo->disp );

   while( 1 ) {
      area_t *ar;

      if( NULLP( timeout_areas ) ) {
	 FD_ZERO( &fd );
	 FD_SET( xfd, &fd );
	 select( xfd + 1, &fd, 0, 0, 0 );
      } else {
	 struct timeval tv;

	 tv.tv_usec = 100000;
	 tv.tv_sec = 0;
	 FD_ZERO( &fd );
	 FD_SET( xfd, &fd );
	 
	 if( select (xfd + 1, &fd, 0, 0, &tv ) == 0 ) {
	    pair_t *lst = timeout_areas;
	    pair_t *prev = 0;

	    while( PAIRP( lst ) ) {
	       area_t *ar = (area_t *)CAR( lst );

	       if( ar->timeout( ar ) ) {
		  prev = lst;
		  lst = CDR( lst );
	       } else {
		  lst = CDR( lst );
		  if( !prev ) {
		     timeout_areas = lst;
		  } else {
		     CDR( prev ) = lst;
		  }
		  free( lst );
	       }
	    }
	 }
      }

      // the evt variable is used to delay Leave/Notify event until
      // no pending events are to be processed
      evt = 0;

      while( XPending( tbar->xinfo->disp ) ) {
	 XNextEvent( tbar->xinfo->disp, &ev );
	 
	 switch( ev.type ) {
	    case ButtonPress:
	       ar = find_area( tbar, ev.xbutton.window );
	       if( ar && ar->button_press ) ar->button_press( &ev, ar );
	       break;

	    case KeyPress:
	       ar = find_area( tbar, ev.xkey.window );
	       if( ar && ar->key_press ) ar->key_press( &ev, ar );
	       break;

	    case DestroyNotify:
	       taskbar_destroy_notify( tbar, &ev );
	       break;

	    case Expose:
	       if( tbar->win == ev.xexpose.window ) {
		  taskbar_refresh( tbar );
	       } else {
		  if( tooltips_windowp( ev.xexpose.window ) ) {
		     tooltips_refresh();
		  } else {
		     ar = find_area( tbar, ev.xexpose.window );
		     if( ar ) ar->refresh( ar );
		  }
	       }
	       break;
	       
	    case PropertyNotify:
	       taskbar_property_notify( tbar, &ev );
	       break;

	    case EnterNotify:
	       evt = EnterNotify;
	       elev = ev;
	       break;

	    case LeaveNotify:
	       evt = LeaveNotify;
	       elev = ev;
	       break;

	    case MotionNotify:
	       ar = find_area( tbar, ev.xmotion.window );
	       iar = enter_area( &ev, ar, iar );
	       break;

	    default:
	       break;
	 }
      }

      // once all the pending events have been handled, we process
      // the last leave/enter event for the tbar window
      switch( evt ) {
	 case EnterNotify:
	    if( tbar->hiddenp ) {
	       taskbar_unhide( tbar );
	    } else {
	       ar = find_area( tbar, ev.xcrossing.window );
	       iar = enter_area( &elev, ar, iar );
	    }
	    
	    break;

	 case LeaveNotify: 
	    iar = enter_area( &elev, 0, iar );
	    
	    tooltips_hide();
	    if( tbar->autohide ) taskbar_hide( tbar );
	    break;
      }
   }
}
