/*=====================================================================*/
/*    serrano/prgm/utils/icccmpanel/mstk.c                             */
/*    -------------------------------------------------------------    */
/*    Author      :  Manuel Serrano                                    */
/*    Creation    :  Sat Oct 11 05:33:42 2003                          */
/*    Last change :  Mon Aug  2 20:55:58 2004 (serrano)                */
/*    Copyright   :  2003-04 Manuel Serrano                            */
/*    -------------------------------------------------------------    */
/*    Small X toolkit                                                  */
/*=====================================================================*/
#include <stdio.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#ifdef HAVE_XPM
#  include <X11/xpm.h>
#endif

#include "config.h"
#include "mstk.h"
#include "tooltips.h"

/*---------------------------------------------------------------------*/
/*    X atoms                                                          */
/*---------------------------------------------------------------------*/
static char *atom_names[] = {
  /* clients */
   "KWM_WIN_ICON",
   "WM_STATE",
   "_MOTIF_WM_HINTS",
   "_NET_WM_STATE",
   "_NET_WM_STATE_SKIP_TASKBAR",
   "_NET_WM_STATE_SHADED",
   "_NET_WM_DESKTOP",
   "_NET_WM_WINDOW_TYPE",
   "_NET_WM_WINDOW_TYPE_DOCK",	/* 8 */
   "_NET_WM_WINDOW_TYPE_TOOLBAR", 
   "_NET_WM_STRUT",
   "_WIN_HINTS",
  /* root */
   "_NET_CLIENT_LIST",
   "_NET_NUMBER_OF_DESKTOPS",
   "_NET_CURRENT_DESKTOP",
   "_NET_WM_STATE_STICKY",
};

#define ATOM_COUNT (sizeof( atom_names ) / sizeof( atom_names[ 0 ]) )
Atom atoms[ ATOM_COUNT ];

/*---------------------------------------------------------------------*/
/*    struct colors                                                    */
/*    static cols ...                                                  */
/*---------------------------------------------------------------------*/
static struct colors {
   unsigned short red, green, blue;
} cols[] = {
   {0xd75c, 0xd75c, 0xd75c},	/* 0. light gray */
   {0xbefb, 0xbaea, 0xbefb},	/* 1. mid gray */
   {0xaefb, 0xaaea, 0xaefb},	/* 2. dark gray */
   {0xefbe, 0xefbe, 0xefbe},	/* 3. white */
   {0x8617, 0x8207, 0x8617},	/* 4. darkest gray */
   {0x0000, 0x0000, 0x0000},	/* 5. black */
   {0xffff, 0xffff, 0xffff},	/* 6  */
   {0xeeee, 0xeeee, 0xeeee},    /* 7  */
   {0xdddd, 0xdddd, 0xdddd},    /* 8  */
   {0xcccc, 0xcccc, 0xcccc},    /* 9  */
   {0xbbbb, 0xbbbb, 0xbbbb},    /* 10 */
   {0xaaaa, 0xaaaa, 0xaaaa},    /* 11 */
   {0x9999, 0x9999, 0x9999},    /* 12 */
   {0x8888, 0x8888, 0x8888},    /* 13 */
   {0x7777, 0x7777, 0x7777},    /* 14 */
   {0x6666, 0x6666, 0x6666},    /* 15 */
   {0x5555, 0x5555, 0x5555},    /* 16 */
   {0x4444, 0x4444, 0x4444},    /* 17 */
   {0x3333, 0x3333, 0x3333},    /* 18 */
   {0x2222, 0x2222, 0x2222},    /* 19 */
   {0x1111, 0x1111, 0x1111},    /* 20 */
   {0x0000, 0x0000, 0x0000},    /* 21 */
   {0xffff, 0x0000, 0x0000},    /* 22 */
   {0xffff, 0xffff, 0xcccc},    /* 23 */
   {0xffff, 0xcccc, 0xffff},    /* 24 */
   {0xcccc, 0xffff, 0xffff},    /* 25 */
   {0x0000, 0x0000, 0xffff},    /* 26 */
   {0x0000, 0xffff, 0x0000},    /* 27 */
};

/*---------------------------------------------------------------------*/
/*    unsigned long                                                    */
/*    palette ...                                                      */
/*---------------------------------------------------------------------*/
#define MSTK_PALETTE_COUNT (sizeof( cols ) / sizeof (cols[ 0 ].red) / 3)
unsigned long mstk_palette[ MSTK_PALETTE_COUNT ];
unsigned long mstk_grey_palette[ 256 ];

/*---------------------------------------------------------------------*/
/*    MSTK_VARS                                                        */
/*---------------------------------------------------------------------*/
#define MSTK_VARS( xinfo ) \
   Display *disp = (xinfo)->disp; \
   GC gc = (xinfo)->gc; \

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    handle_error ...                                                 */
/*---------------------------------------------------------------------*/
static void
handle_error( Display * d, XErrorEvent * ev ) {
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    init_X ...                                                       */
/*---------------------------------------------------------------------*/
static int
init_X( Xinfo_t *xinfo ) {
   Display *disp;
   int screen;
   XGCValues gcv;
   XGCValues gcpv;
    
   if( !(disp = XOpenDisplay( NULL )) )
      return 1;

   xinfo->disp = disp;
   xinfo->screen = screen = DefaultScreen( disp );
   xinfo->screen_depth = DefaultDepth( disp, screen );
   xinfo->screen_height = DisplayHeight( disp, screen );
   xinfo->screen_width = DisplayWidth( disp, screen );
   xinfo->root_win = RootWindow( disp, screen );

   /* helps us catch windows closing/opening */
   XSelectInput( disp, xinfo->root_win, PropertyChangeMask );

   XSetErrorHandler( (XErrorHandler)handle_error );

   do {
      xinfo->xfsb = XLoadQueryFont( disp, xinfo->fontname_bold );
      xinfo->fontname_bold = "fixed";
   } while( !xinfo->xfsb );

   do {
      xinfo->xfsp = XLoadQueryFont( disp, xinfo->fontname_plain );
      xinfo->fontname_plain = "fixed";
   } while( !xinfo->xfsp );

   gcv.graphics_exposures = False;
   xinfo->text_y = xinfo->xfsb->ascent + ((20 - xinfo->xfsb->ascent) / 2);
   gcv.font = xinfo->xfsb->fid;
   xinfo->gc = XCreateGC( disp,
			  xinfo->root_win,
			  GCFont | GCGraphicsExposures,
			  &gcv );
   
   gcpv.graphics_exposures = False;
   gcpv.font = xinfo->xfsp->fid;
   xinfo->gcp = XCreateGC( disp,
			   xinfo->root_win,
			   GCFont | GCGraphicsExposures,
			   &gcpv );
   
   return 0;
}

/*---------------------------------------------------------------------*/
/*    Xinfo_t *                                                        */
/*    init_mstk ...                                                    */
/*---------------------------------------------------------------------*/
Xinfo_t *
init_mstk( config_t *config ) {
   XColor xcl;
   unsigned int i;
   Xinfo_t *xinfo = (Xinfo_t *)calloc( 1, sizeof( Xinfo_t ) );
   
   xinfo->fontname_plain = config->fontname_plain;
   xinfo->fontname_bold = config->fontname_bold;
   xinfo->tooltips_fontname = config->tooltips_fontname;

   /* initialize X connection */
   if( init_X( xinfo ) )
      return 0L;

   /* initialize colors */
   for( i = 0; i < MSTK_PALETTE_COUNT; i++ ) {
      xcl.red = cols[ i ].red;
      xcl.blue = cols[ i ].blue;
      xcl.green = cols[ i ].green;
      XAllocColor( xinfo->disp,
		   DefaultColormap( xinfo->disp, xinfo->screen ),
		   &xcl );
      mstk_palette[ i ] = xcl.pixel;
   }

   for( i = 0; i < 256; i++ ) {
      xcl.red = xcl.blue = xcl.green = i | (i<<8);
      XAllocColor( xinfo->disp,
		   DefaultColormap( xinfo->disp, xinfo->screen ),
		   &xcl );
      mstk_grey_palette[ i ] = xcl.pixel;
   }

   XInternAtoms( xinfo->disp, atom_names, ATOM_COUNT, False, atoms );

   /* initialize the tooltips */
   init_tooltips( xinfo );
   
   return xinfo;
}

/*---------------------------------------------------------------------*/
/*    Xinfo_t *                                                        */
/*    copy_xinfo ...                                                   */
/*---------------------------------------------------------------------*/
Xinfo_t *
copy_xinfo( Xinfo_t *old ) {
   Xinfo_t *new = (Xinfo_t *)calloc( 1, sizeof( Xinfo_t ) );

   *new = *old;

   return new;
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    closeup_mstk ...                                                 */
/*---------------------------------------------------------------------*/
void
closeup_mstk( Xinfo_t *xinfo ) {
   XCloseDisplay( xinfo->disp );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    scale_icon ...                                                   */
/*---------------------------------------------------------------------*/
void
scale_icon( Xinfo_t *xinfo, Window win, 
	    Pixmap icon, Pixmap mask,
	    Pixmap *ricon, Pixmap *rmask,
	    int nw, int nh ) {
   MSTK_VARS( xinfo );
   int depth = xinfo->screen_depth;
   int xx, yy, x, y, w, h, d, bw;
   Pixmap pix, mk = None;
   XGCValues gcv;
   GC mgc;

   XGetGeometry( disp, icon, &pix, &x, &y, &w, &h, &bw, &d );
   pix = XCreatePixmap( disp, win, nw, nh, depth );

   if (mask != None) {
      mk = XCreatePixmap( disp, win, nw, nw, 1 );
      gcv.subwindow_mode = IncludeInferiors;
      gcv.graphics_exposures = False;
      mgc = XCreateGC( disp, mk, GCGraphicsExposures | GCSubwindowMode, &gcv );
   }

   /* this is my simple & dirty scaling routine */
   for( y = nh - 1; y >= 0; y-- ) {
      yy = (y * h) / nh;
      for( x = nw - 1; x >= 0; x-- ) {
	 xx = (x * w) / nw;
	 if( d != depth )
	    XCopyPlane( disp, icon, pix, gc, xx, yy, 1, 1, x, y, 1 );
	 else
	    XCopyArea( disp, icon, pix, gc, xx, yy, 1, 1, x, y );
	 if( mk != None )
	    XCopyArea( disp, mask, mk, mgc, xx, yy, 1, 1, x, y );
      }
   }

   if( mk != None ) {
      XFreeGC( disp, mgc );
      *rmask = mk;
   }

   *ricon = pix;
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    client_msg ...                                                   */
/*---------------------------------------------------------------------*/
static int
client_msg( Display * disp, Window win, char *msg,
	    unsigned long data0, unsigned long data1,
	    unsigned long data2, unsigned long data3, unsigned long data4 ) {
   XEvent event;
   long mask = SubstructureRedirectMask | SubstructureNotifyMask;

   event.xclient.type = ClientMessage;
   event.xclient.serial = 0;
   event.xclient.send_event = True;
   event.xclient.message_type = XInternAtom( disp, msg, False );
   event.xclient.window = win;
   event.xclient.format = 32;
   event.xclient.data.l[ 0 ] = data0;
   event.xclient.data.l[ 1 ] = data1;
   event.xclient.data.l[ 2 ] = data2;
   event.xclient.data.l[ 3 ] = data3;
   event.xclient.data.l[ 4 ] = data4;

   if( XSendEvent(disp, DefaultRootWindow(disp), False, mask, &event) ) {
      return 0;
   } else {
      fprintf( stderr, "Cannot send %s event.\n", msg );
      return 1;
   }
}

/*---------------------------------------------------------------------*/
/*    void *                                                           */
/*    get_window_prop_data ...                                         */
/*---------------------------------------------------------------------*/
void *
get_window_prop_data( Display *disp, Window win,
		      Atom prop, Atom type, int *items ) {
   Atom type_ret;
   int format_ret;
   unsigned long items_ret;
   unsigned long after_ret;
   unsigned char *prop_data = 0;

   XGetWindowProperty( disp, win, prop, 0, 0x7fffffff, False,
		       type,
		       &type_ret, &format_ret, &items_ret,
		       &after_ret, &prop_data );

   if( items ) *items = items_ret;

   return prop_data;
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    get_window_prop_int ...                                          */
/*---------------------------------------------------------------------*/
static int
get_window_prop_int( Display *disp, Window win, Atom at ) {
   int num = 0;
   unsigned long *data;

   data = get_window_prop_data( disp, win, at, XA_CARDINAL, 0 );
   if( data ) {
      num = *data;
      XFree( data );
   }
   return num;
}

/*---------------------------------------------------------------------*/
/*    char *                                                           */
/*    window_name ...                                                  */
/*---------------------------------------------------------------------*/
char *
window_name( Display *disp, Window win ) {
   return get_window_prop_data( disp, win, XA_WM_NAME, XA_STRING, 0 );
}

/*---------------------------------------------------------------------*/
/*    char *                                                           */
/*    window_class ...                                                 */
/*---------------------------------------------------------------------*/
char *
window_class( Display *disp, Window win ) {
   int num;
   XClassHint ch;

   if( XGetClassHint( disp, win, &ch ) == 0) {
      return get_window_prop_data( disp, win, XA_WM_CLASS, XA_STRING, &num );
   } else {
      char *res = 0;
      if( ch.res_class != 0 ) {
	 res = strdup( ch.res_class );
	 XFree( ch.res_class );
      }

      if( ch.res_name != 0 )
	 XFree( ch.res_name );

      return res;
   }
}

/*---------------------------------------------------------------------*/
/*    char                                                             */
/*    window_hint_icon ...                                             */
/*---------------------------------------------------------------------*/
char
window_hint_icon( Display *disp, Window win, Pixmap *icon, Pixmap *mask ) {
   XWMHints *hin;

   hin = (XWMHints *)get_window_prop_data( disp, win, XA_WM_HINTS, XA_WM_HINTS, 0 );

   if( hin ) {
      int res = 0;
      
      if( (hin->flags & IconPixmapHint) ) {
	 if( (hin->flags & IconMaskHint) ) {
	    *mask = hin->icon_mask;
	 }

	 *icon = hin->icon_pixmap;
	 res = 1;
      }
      XFree( hin );

      return res;
   } else {
      return 0;
   }
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    set_window_prop ...                                              */
/*---------------------------------------------------------------------*/
void
set_window_prop( Display *disp,  Window win, Atom at, Atom type, long val ) {
   XChangeProperty( disp, win, at, type, 32,
		    PropModeReplace, (unsigned char *)&val, 1 );
}

/*---------------------------------------------------------------------*/
/*    Window *                                                         */
/*    desktop_windows ...                                              */
/*---------------------------------------------------------------------*/
Window *
desktop_windows( Xinfo_t *xinfo, int *num ) {
   get_window_prop_data( xinfo->disp, xinfo->root_win,
			 atom__NET_CLIENT_LIST, XA_WINDOW,
			 num );
}

/*---------------------------------------------------------------------*/
/*    int                                                              */
/*    window_desktop ...                                               */
/*---------------------------------------------------------------------*/
int
window_desktop( Display *disp, Window win ) {
   return get_window_prop_int( disp, win, atom__NET_WM_DESKTOP );
}

/*---------------------------------------------------------------------*/
/*    int                                                              */
/*    current_desktop ...                                              */
/*---------------------------------------------------------------------*/
int
current_desktop( Display *disp, Window win ) {
   return get_window_prop_int( disp, win, atom__NET_CURRENT_DESKTOP );
}

/*---------------------------------------------------------------------*/
/*    int                                                              */
/*    number_of_desktops ...                                           */
/*---------------------------------------------------------------------*/
int
number_of_desktops( Display *disp, Window win ) {
   return get_window_prop_int( disp, win, atom__NET_NUMBER_OF_DESKTOPS );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    switch_desktop ...                                               */
/*---------------------------------------------------------------------*/
void
switch_desktop( Display *disp, Window win, int desk ) {
   XClientMessageEvent xev;

   xev.type = ClientMessage;
   xev.window = win;
   xev.message_type = atom__NET_CURRENT_DESKTOP;
   xev.format = 32;
   xev.data.l[ 0 ] = desk;
   XSendEvent( disp, win, False, SubstructureNotifyMask, (XEvent *)&xev );
}

/*---------------------------------------------------------------------*/
/*    int                                                              */
/*    window_iconifiedp ...                                            */
/*---------------------------------------------------------------------*/
int
window_iconifiedp( Display *disp, Window win ) {
   unsigned long *data;
   int ret = 0;
   
   data = get_window_prop_data( disp, win, atom_WM_STATE, atom_WM_STATE, 0 );

   if( data ) {
      if( data[ 0 ] == IconicState ) 
	 ret = 1;
      XFree( data );
   }
   
   return ret;
}

/*---------------------------------------------------------------------*/
/*    int                                                              */
/*    window_hiddenp ...                                               */
/*---------------------------------------------------------------------*/
int
window_hiddenp( Display *disp, Window win ) {
   unsigned long *data;
   int ret = 0;
   int num;

   data = get_window_prop_data( disp, win, atom__NET_WM_STATE, XA_ATOM, &num );
   
   if( data ) {
      while( num ) {
	 num--;
	 if( (data[num]) == atom__NET_WM_STATE_SKIP_TASKBAR )
	    ret = 1;
      }
      XFree( data );
   }

   return ret;
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    window_deiconify ...                                             */
/*---------------------------------------------------------------------*/
void
window_deiconify( Display *disp, Window win ) {
   client_msg( disp, win, "_NET_ACTIVE_WINDOW", 0, 0, 0, 0, 0 );
   XMapRaised( disp, win );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    draw_text ...                                                    */
/*---------------------------------------------------------------------*/
void
draw_text( Xinfo_t *xinfo, Window win, int x, int y,
	   char *text, int len,
	   unsigned long palette[], int c ) {
   MSTK_VARS( xinfo );
   unsigned long *color = palette ? palette : mstk_palette;
   XSetForeground( disp, gc, color[ c ] );
   
   XDrawString( disp, win, gc, x, y, text, len );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    draw_text_plain ...                                              */
/*---------------------------------------------------------------------*/
void
draw_text_plain( Xinfo_t *xinfo, Window win, int x, int y,
	   char *text, int len,
	   unsigned long palette[], int c ) {
   Display *disp = (xinfo)->disp;
   GC gc = (xinfo)->gcp;
   unsigned long *color = palette ? palette : mstk_palette;
   XSetForeground( disp, gc, color[ c ] );
   
   XDrawString( disp, win, gc, x, y, text, len );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    draw_line ...                                                    */
/*---------------------------------------------------------------------*/
void
draw_line( Xinfo_t *xinfo, Window win,
	   int x, int y, int w , int h,
	   unsigned long palette[], int c ) {
   MSTK_VARS( xinfo );
   unsigned long *color = palette ? palette : mstk_palette;
   XSetForeground( disp, gc, color[ c ] );
   XDrawLine( disp, win, gc, x, y, x + w, y + h );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    draw_gradient ...                                                */
/*    -------------------------------------------------------------    */
/*    Fill a window with a gradient coloring starting from COLOR       */
/*    CF to CL.                                                        */
/*---------------------------------------------------------------------*/
void
draw_gradient( Xinfo_t *xinfo, Window win,
	       int x, int y, int w, int h,
	       unsigned long palette[], int cf, int step, int sw ) {
   MSTK_VARS( xinfo );
   int i, c;
   unsigned long *color = palette ? palette : mstk_palette;

   if( step ) {
      for( i = 0, c = cf; i < h; i++ ) {
	 XSetForeground( disp, gc, color[ c ] );
	 XDrawLine( disp, win, gc, x, y + i, x + w, y + i );

	 if( !(i % sw) ) c += step;
      }
   } else {
      for( i = 0; i < h; i++ ) {
	 XSetForeground( disp, gc, color[ cf ] );
	 XDrawLine( disp, win, gc, x, y + i, x + w, y + i );
      }
   }
}


/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    draw_relief ...                                                  */
/*---------------------------------------------------------------------*/
void
draw_relief( Xinfo_t *xinfo, Window win, 
	     int x, int y, int w, int h,
	     unsigned long palette[], int c0, int c1,
	     int border ) {
   MSTK_VARS( xinfo );
   unsigned long *color = palette ? palette : mstk_palette;
   int i;
   
   for( i = 0; i < border; i++ ) {
      XSetForeground( disp, gc, color[ c0 ] );
      XDrawLine( disp, win, gc, x + i, y + i, x + i, y + h - i );
      XDrawLine( disp, win, gc, x + i, y + i, x + w - i, y + i );
      XSetForeground( disp, gc, color[ c1 ] );
      XDrawLine( disp, win, gc, x + w - i , y + h - i, x + w - i, y + i );
      XDrawLine( disp, win, gc, x + i + 1, y + h - i, x + w - i, y + h - i );
   }
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    draw_partial_relief ...                                          */
/*---------------------------------------------------------------------*/
void
draw_partial_relief( Xinfo_t *xinfo, Window win, int mask,
		     int x, int y, int w, int h,
		     unsigned long palette[], int c0, int c1,
		     int border ) {
   MSTK_VARS( xinfo );
   unsigned long *color = palette ? palette : mstk_palette;
   int i;
   
   for( i = 0; i < border; i++ ) {
      XSetForeground( disp, gc, color[ c0 ] );
      if( mask & RELIEF_LEFT )
	 XDrawLine( disp, win, gc, x + i, y + i, x + i, y + h - i );
      if( mask & RELIEF_TOP )
	 XDrawLine( disp, win, gc, x + i, y + i, x + w - i, y + i );
      XSetForeground( disp, gc, color[ c1 ] );
      if( mask & RELIEF_RIGHT )
	 XDrawLine( disp, win, gc, x + w - i , y + h - i, x + w - i, y + i );
      if( mask & RELIEF_BOTTOM )
	 XDrawLine( disp, win, gc,
		    x + i + (mask & RELIEF_LEFT) ? 1 : 0, y + h - i,
		    x + w - i, y + h - i );

   }
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    draw_pixmap ...                                                  */
/*---------------------------------------------------------------------*/
void
draw_pixmap( Xinfo_t *xinfo, Window win,
	     Pixmap icon, Pixmap mask,
	     int x, int y, int w, int h ) {
   MSTK_VARS( xinfo );
   XSetClipMask( disp, gc, mask );
   XSetClipOrigin( disp, gc, x, y );
   XCopyArea( disp, icon, win, gc, 0, 0, w, h, x, y );
   XSetClipMask( disp, gc, None );
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    draw_dot ...                                                     */
/*---------------------------------------------------------------------*/
static void
draw_dot( Display *disp, Window win, GC gc, int x, int y ) {
   XSetForeground( disp, gc, mstk_grey_palette[ 0xf7 ] );
   XDrawPoint( disp, win, gc, x, y );
   XSetForeground( disp, gc, mstk_grey_palette[ 0x40 ] );
   XDrawPoint( disp, win, gc, x + 1, y + 1 );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    draw_grill ...                                                   */
/*---------------------------------------------------------------------*/
void
draw_grill( Xinfo_t *xinfo, Window win, int x, int y, int w, int h ) {
   MSTK_VARS( xinfo );
   int xx;

   for( xx = 2; xx < w; xx += 3 ) {
      int yy = y + 1;

      while( yy <= h ) {
	 draw_dot( disp, win, gc, x + xx, yy );
	 yy += 3;
      }
   }
}


